@@ -28,14 +28,17 @@ class UserAdmin(admin.ModelAdmin):
     ]
     readonly_fields = ["keycloak_link", "email", "first_name", "last_name"]
 
-    search_fields = ['username', 'email']
+    search_fields = ["username", "email"]
 
     def keycloak_link(self, obj):
         config = settings.KEYCLOAK_CONFIG
         label = obj.id
         link = KEYCLOAK_ADMIN_USER_PAGE.format(
-            host=config.get("SERVER_URL"), realm=config.get("REALM"), id=label)
-        return format_html('<a href="{link}" target="_blank">{label}</a>', link=link, label=label)
+            host=config.get("SERVER_URL"), realm=config.get("REALM"), id=label
+        )
+        return format_html(
+            '<a href="{link}" target="_blank">{label}</a>', link=link, label=label
+        )
 
     keycloak_link.short_description = _("keycloak link")
 
